/*
    -- MAGMA (version 1.5.0-beta1-sparse) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date May 2014

       @author Hartwig Anzt 

       @generated from zbailu.cpp normal z -> c, Thu May  1 11:45:22 2014
*/
// includes CUDA
#include <cuda_runtime_api.h>
#include <cublas.h>
#include <cusparse_v2.h>
#include <cuda_profiler_api.h>

// project includes
#include "common_magma.h"
#include "../include/magmasparse.h"

#include <assert.h>


#define PRECISION_c


/*  -- MAGMA (version 1.5.0-beta1-sparse) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date May 2014

    Purpose
    =======

    Prepares the ILU preconditioner via the asynchronous ILU iteration.

    Arguments
    =========

    magma_c_sparse_matrix A                   input matrix A
    magma_c_preconditioner *precond           preconditioner parameters

    ========================================================================  */

magma_int_t
magma_cailusetup( magma_c_sparse_matrix A, magma_c_preconditioner *precond ){


    magma_c_sparse_matrix hA, hAL, hAU, dAL, dAU, hL, hU, 
                                        dL, dU, DL, RL, DU, RU;

    // copy original matrix as CSRCSC to device
    magma_c_mtransfer(A, &hA, A.memory_location, Magma_CPU);
    hAL.storage_type = Magma_CSRCSCL;
    magma_c_mconvert( hA, &hAL, Magma_CSR, hAL.storage_type );
    hAU.storage_type = Magma_CSRCSCU;
    magma_c_mconvert( hA, &hAU, Magma_CSR, hAU.storage_type );

    magma_c_mtransfer( hAL, &dAL, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( hAU, &dAU, Magma_CPU, Magma_DEV );

    magma_c_mfree(&hAL);
    magma_c_mfree(&hAU);


    // scale initial guess
    #ifdef PRECISION_d
    for(magma_int_t z=0; z<hA.num_rows; z++){
        real_Double_t s = 0.0;
        for(magma_int_t f=hA.row[z]; f<hA.row[z+1]; f++)
            s+= MAGMA_C_REAL(hA.val[f])*MAGMA_C_REAL(hA.val[f]);
        s = 1.0/sqrt(   s  );
        for(magma_int_t f=hA.row[z]; f<hA.row[z+1]; f++)
            hA.val[f] = hA.val[f] * s ;
           
    }
    #endif

    hAL.storage_type = Magma_CSRCSCL;
    magma_c_mconvert( hA, &hAL, Magma_CSR, hAL.storage_type );
    hAU.storage_type = Magma_CSRCSCU;
    magma_c_mconvert( hA, &hAU, Magma_CSR, hAU.storage_type );

    magma_c_mfree(&hA);

    magma_c_mtransfer( hAL, &dL, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( hAU, &dU, Magma_CPU, Magma_DEV );

    magma_c_mfree(&hAL);
    magma_c_mfree(&hAU);

    for(int i=0; i<3; i++){
        magma_cailu_csr_s( dAL, dAU, dL, dU );

    }
    magma_c_mtransfer( dL, &precond->M, Magma_DEV, Magma_DEV );
    magma_c_mtransfer( dL, &hL, Magma_DEV, Magma_CPU );
    magma_c_mtransfer( dU, &hU, Magma_DEV, Magma_CPU );

    magma_c_mfree(&dL);
    magma_c_mfree(&dU);
    magma_c_mfree(&dAL);
    magma_c_mfree(&dAU);


    magma_c_mconvert(hL, &hAL, hL.storage_type, Magma_CSR);
    magma_c_mconvert(hU, &hAU, hU.storage_type, Magma_CSR);

    magma_c_mfree(&hL);
    magma_c_mfree(&hU);

    magma_ccsrsplit( 256, hAL, &DL, &RL );
    magma_ccsrsplit( 256, hAU, &DU, &RU );

    magma_c_mfree(&hAL);
    magma_c_mfree(&hAU);

    magma_c_mtransfer( DL, &precond->LD, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( DU, &precond->UD, Magma_CPU, Magma_DEV );

    magma_c_mtransfer( RL, &precond->L, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( RU, &precond->U, Magma_CPU, Magma_DEV );

    magma_c_mfree(&DL);
    magma_c_mfree(&RL);
    magma_c_mfree(&DU);
    magma_c_mfree(&RU);

    return MAGMA_SUCCESS;

}


/*  -- MAGMA (version 1.5.0-beta1-sparse) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date May 2014

    Purpose
    =======

    Performs the left triangular solves using the ILU preconditioner.

    Arguments
    =========

    magma_c_vector b                        RHS
    magma_c_vector *x                       vector to precondition
    magma_c_preconditioner *precond         preconditioner parameters

    ========================================================================  */

magma_int_t
magma_capplyailu_l( magma_c_vector b, magma_c_vector *x, 
                    magma_c_preconditioner *precond ){

    magma_int_t iters = 1;
    for(int k=0; k<15; k++)
        magma_cbajac_csr( iters, precond->LD, precond->L, b, x );
           
    return MAGMA_SUCCESS;

}


/*  -- MAGMA (version 1.5.0-beta1-sparse) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date May 2014

    Purpose
    =======

    Performs the right triangular solves using the ILU preconditioner.

    Arguments
    =========

    magma_c_vector b                        RHS
    magma_c_vector *x                       vector to precondition
    magma_c_preconditioner *precond         preconditioner parameters

    ========================================================================  */

magma_int_t
magma_capplyailu_r( magma_c_vector b, magma_c_vector *x, 
                    magma_c_preconditioner *precond ){

    magma_int_t iters = 1;
    for(int k=0; k<15; k++)
        magma_cbajac_csr( iters, precond->UD, precond->U, b, x );

    return MAGMA_SUCCESS;

}





/*  -- MAGMA (version 1.5.0-beta1-sparse) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date May 2014

    Purpose
    =======

    Prepares the IC preconditioner via the asynchronous IC iteration.

    Arguments
    =========

    magma_c_sparse_matrix A                   input matrix A
    magma_c_preconditioner *precond           preconditioner parameters

    ========================================================================  */

magma_int_t
magma_caiccsetup( magma_c_sparse_matrix A, magma_c_preconditioner *precond ){


    magma_c_sparse_matrix hA, hAL, dAL, hL, dL, DL, RL;

    // copy original matrix as CSRCSC to device
    magma_c_mtransfer(A, &hA, A.memory_location, Magma_CPU);

    hAL.storage_type = Magma_CSRCSCL;
    magma_c_mconvert( hA, &hAL, Magma_CSR, hAL.storage_type );
    magma_c_mtransfer( hAL, &dAL, Magma_CPU, Magma_DEV );
    magma_c_mfree(&hAL);

    // scale initial guess
    #ifdef PRECISION_d
    for(magma_int_t z=0; z<hA.num_rows; z++){
        real_Double_t s = 0.0;
        for(magma_int_t f=hA.row[z]; f<hA.row[z+1]; f++)
            s+= MAGMA_C_REAL(hA.val[f])*MAGMA_C_REAL(hA.val[f]);
        s = 1.0/sqrt(   s  );
        for(magma_int_t f=hA.row[z]; f<hA.row[z+1]; f++)
            hA.val[f] = hA.val[f] * s ;
           
    }
    #endif

    hAL.storage_type = Magma_CSRCSCL;
    magma_c_mconvert( hA, &hAL, Magma_CSR, hAL.storage_type );
    magma_c_mtransfer( hAL, &dL, Magma_CPU, Magma_DEV );
    magma_c_mfree(&hAL);
    magma_c_mfree(&hA);

    for(int i=0; i<3; i++){
        magma_caic_csr_s( dAL, dL );

    }
    magma_c_mtransfer( dL, &hL, Magma_DEV, Magma_CPU );

    magma_c_mfree(&dL);
    magma_c_mfree(&dAL);

    magma_c_mconvert(hL, &hAL, hL.storage_type, Magma_CSR);

    magma_c_mtransfer( hAL, &precond->M, Magma_CPU, Magma_DEV );

    magma_ccsrsplit( 256, hAL, &DL, &RL );

    magma_c_mtransfer( DL, &precond->UD, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( RL, &precond->U, Magma_CPU, Magma_DEV );

    magma_c_mfree(&hL);

    magma_c_cucsrtranspose(   hAL, &hL );

    magma_ccsrsplit( 256, hL, &DL, &RL );

    magma_c_mtransfer( DL, &precond->LD, Magma_CPU, Magma_DEV );
    magma_c_mtransfer( RL, &precond->L, Magma_CPU, Magma_DEV );

    magma_c_mfree(&hAL);
    magma_c_mfree(&hL);

    magma_c_mfree(&DL);
    magma_c_mfree(&RL);


    // CUSPARSE context //
    cusparseHandle_t cusparseHandle;
    cusparseStatus_t cusparseStatus;
    cusparseStatus = cusparseCreate(&cusparseHandle);
     if(cusparseStatus != 0)    printf("error in Handle.\n");

    cusparseMatDescr_t descrL;
    cusparseStatus = cusparseCreateMatDescr(&descrL);
     if(cusparseStatus != 0)    printf("error in MatrDescr.\n");

    cusparseStatus =
    cusparseSetMatType(descrL,CUSPARSE_MATRIX_TYPE_TRIANGULAR);
     if(cusparseStatus != 0)    printf("error in MatrType.\n");

    cusparseStatus =
    cusparseSetMatDiagType (descrL, CUSPARSE_DIAG_TYPE_UNIT);
     if(cusparseStatus != 0)    printf("error in DiagType.\n");

    cusparseStatus =
    cusparseSetMatIndexBase(descrL,CUSPARSE_INDEX_BASE_ZERO);
     if(cusparseStatus != 0)    printf("error in IndexBase.\n");

    cusparseStatus =
    cusparseSetMatFillMode(descrL,CUSPARSE_FILL_MODE_LOWER);
     if(cusparseStatus != 0)    printf("error in fillmode.\n");


    cusparseStatus = cusparseCreateSolveAnalysisInfo(&precond->cuinfoL); 
     if(cusparseStatus != 0)    printf("error in info.\n");

    cusparseStatus =
    cusparseCcsrsv_analysis(cusparseHandle, 
        CUSPARSE_OPERATION_TRANSPOSE, precond->M.num_rows, 
        precond->M.nnz, descrL, 
        precond->M.val, precond->M.row, precond->M.col, precond->cuinfoL );
     if(cusparseStatus != 0)    printf("error in analysis L.\n");

    cusparseDestroyMatDescr( descrL );

    cusparseMatDescr_t descrU;
    cusparseStatus = cusparseCreateMatDescr(&descrU);
     if(cusparseStatus != 0)    printf("error in MatrDescr.\n");

    cusparseStatus =
    cusparseSetMatType(descrU,CUSPARSE_MATRIX_TYPE_TRIANGULAR);
     if(cusparseStatus != 0)    printf("error in MatrType.\n");

    cusparseStatus =
    cusparseSetMatDiagType (descrU, CUSPARSE_DIAG_TYPE_NON_UNIT);
     if(cusparseStatus != 0)    printf("error in DiagType.\n");

    cusparseStatus =
    cusparseSetMatIndexBase(descrU,CUSPARSE_INDEX_BASE_ZERO);
     if(cusparseStatus != 0)    printf("error in IndexBase.\n");

    cusparseStatus =
    cusparseSetMatFillMode(descrU,CUSPARSE_FILL_MODE_LOWER);
     if(cusparseStatus != 0)    printf("error in fillmode.\n");

    cusparseStatus = cusparseCreateSolveAnalysisInfo(&precond->cuinfoU); 
     if(cusparseStatus != 0)    printf("error in info.\n");

    cusparseStatus =
    cusparseCcsrsv_analysis(cusparseHandle, 
        CUSPARSE_OPERATION_NON_TRANSPOSE, precond->M.num_rows, 
        precond->M.nnz, descrU, 
        precond->M.val, precond->M.row, precond->M.col, precond->cuinfoU );
     if(cusparseStatus != 0)    printf("error in analysis U.\n");

    cusparseDestroyMatDescr( descrU );
    cusparseDestroy( cusparseHandle );

    return MAGMA_SUCCESS;

}
